use std::rc::{Rc, Weak};
use std::cell::RefCell;
use std::any::Any;

const TVN_BITS: usize = 6;
const TVR_BITS: usize = 8;
const TVN_SIZE: usize = 1 << TVN_BITS;
const TVR_SIZE: usize  = 1 << TVR_BITS;
const TVN_MASK: usize = TVN_SIZE - 1;
const TVR_MASK: usize = TVR_SIZE - 1;

pub type TimerAttacherRp = Rc::<RefCell::<TimerAttacher>>;
pub type TimerWp = Weak::<Rc::<RefCell::<TimerList>>>;
pub type TimerCb = Box::<dyn FnMut (TimerEvent, Option<Vec<Rc::<dyn Any>>>)>;
type TimerListListHeadRp = Rc::<RefCell::<crate::ListHead<TimerList>>>;
type TimerListRp = Rc::<RefCell::<TimerList>>;
type TimerListWp = Weak::<RefCell::<TimerList>>;

pub enum TimerEvent {
    Exec,
    Delete,
    Destory,
    Custom(i32)
}

pub struct TimerAttacher {
    timers: TimerListListHeadRp,
}

enum TimerType {
    Timeout,
    Interval
}

impl Default for TimerType {

    fn default() -> Self
    {
        return TimerType::Timeout;
    }

}

#[derive(Default)]
pub struct TimerList {
    holder: TimerListRp,
    wp: Rc::<TimerListRp>,
    timer_entry: TimerListListHeadRp,
    attach_entry: TimerListListHeadRp,
    timer_type: TimerType,
    expire_time: i32,
    expires: i64,

    cb: Option<TimerCb>,
}

pub struct Timer {
    free_timer_num: i32,
    free_timer_list: TimerListListHeadRp,
    running_timer: TimerListWp,
    timer_tick: i64,
    get_tick_count: fn () -> i64,
    cache_timer_num: i32,
    work_list: TimerListListHeadRp,
    tv1: [TimerListListHeadRp; TVR_SIZE],
    tv2: [TimerListListHeadRp; TVN_SIZE],
    tv3: [TimerListListHeadRp; TVN_SIZE],
    tv4: [TimerListListHeadRp; TVN_SIZE],
    tv5: [TimerListListHeadRp; TVN_SIZE],
}

impl Timer {

    pub fn init(mut self,
                get_tick_count: fn () -> i64,
                gctime: i32,
                cache_timer_num: i32) {
        self.get_tick_count = get_tick_count;
        self.cache_timer_num = cache_timer_num;
        self.free_timer_num = 0;
        self.work_list = crate::ListHead::<TimerList>::new_head();
        for i in 0..self.tv1.len() {
            self.tv1[i] = crate::ListHead::<TimerList>::new_head();
        }
        for i in 0..self.tv2.len() {
            self.tv2[i] = crate::ListHead::<TimerList>::new_head();
        }
        for i in 0..self.tv3.len() {
            self.tv3[i] = crate::ListHead::<TimerList>::new_head();
        }
        for i in 0..self.tv4.len() {
            self.tv4[i] = crate::ListHead::<TimerList>::new_head();
        }
        for i in 0..self.tv5.len() {
            self.tv5[i] = crate::ListHead::<TimerList>::new_head();
        }
        self.timer_tick = (self.get_tick_count)();
        let self_wp = self.shared_from_self();
        let cb = Box::new(
            move |_event: TimerEvent, _args: Option<Vec<Rc::<dyn Any>>>| {
                self_wp.borrow_mut().gc_timer(_event, _args);
            }
        );

        self.set_interval(
            gctime,
            cb,
        );
    }

    fn shared_from_self(&self) -> Rc::<RefCell::<Timer>> {
        //let cell_rc: &Rc<RefCell<Timer> = &self ;
        //let head: &ListHead<T> = &*cell_rc.borrow();
        let cell_start = 0 as *const u8;
        //let data_start = head as *const ListHead<T> as *const u8;
        //let cell_offset = unsafe { data_start.offset_from(cell_start) };
        unsafe {
            return Rc::from_raw(cell_start as *const RefCell::<Timer>);
        };
    }

    pub fn uninit(&mut self) {
        self.clear();
    }

    fn clear(&mut self) {

    }

    fn new_timer_list(&mut self) -> Rc::<RefCell::<TimerList>> {
        let p: Rc::<RefCell::<TimerList>>;
        if self.free_timer_list.borrow().empty() {
            p = self.free_timer_list.borrow().first_entry().upgrade().unwrap();
            if Rc::weak_count(&p.borrow().wp) > 0 {
                p.borrow_mut().wp = Rc::new(p.borrow().holder.clone());
            }
        } else {
            p = Rc::new(RefCell::new(TimerList{
                holder: Default::default(),
                wp: Default::default(),
                timer_entry: Default::default(),
                attach_entry: Default::default(),
                timer_type: TimerType::Timeout,
                expire_time: 0,
                expires: 0,
                cb: None,
            }));
            p.borrow_mut().holder = p.clone();
            p.borrow_mut().wp = Rc::new(p.borrow().holder.clone());
            p.borrow_mut().timer_entry = crate::ListHead::<TimerList>::new_node(
                Rc::downgrade(&p.borrow().holder)
            );
            p.borrow_mut().attach_entry = crate::ListHead::<TimerList>::new_head();
        }
        return p;
    }

    pub fn update(&mut self) {
        while (self.get_tick_count)() >= self.timer_tick {
            let index = (self.timer_tick & (TVR_MASK as i64)) as usize;
            if index == 0 {

            }
            self.timer_tick += 1;

            crate::ListHead::<TimerList>::replace_init
                (
                    &mut self.work_list,
                    &mut self.tv1[index]
                );
            while !self.work_list.borrow().empty() {
                let timer = &self.work_list.borrow().first_entry();
                self.running_timer = timer.clone();
                match &timer.upgrade().unwrap().borrow_mut().cb {
                    Some(_v) => {
                        //(*v)(0, None);
                    }
                    None => {

                    }
                }
            }
        }
    }

    fn internal_add(&mut self,
                    timer_type: TimerType,
                    expire_time: i32,
                    cb: TimerCb,
                    attacher: Option<TimerAttacherRp>) -> TimerWp {
        let t = self.new_timer_list();
        t.borrow_mut().cb = Some(cb);
        t.borrow_mut().timer_type = timer_type;
        t.borrow_mut().expire_time = expire_time;
        t.borrow_mut().expires = (self.get_tick_count)() + expire_time as i64;
        match attacher {
            Some(v) => {
                crate::ListHead::add_tail(&v.borrow_mut().timers,
                                          &t.borrow_mut().attach_entry);
            }
            None => {

            }
        }
        {
            let idx = t.borrow().expires - self.timer_tick;
            let index;
            let vec: &mut Rc::<RefCell::<crate::ListHead<TimerList>>>;
            if idx < 0 {
                index = self.timer_tick & (TVR_MASK as i64);
                vec = &mut self.tv1[index as usize];
            } else if idx < TVR_SIZE as i64 {
                index = t.borrow().expires & (TVR_MASK as i64);
                vec = &mut self.tv1[index as usize];
            } else if idx < (1 << (TVR_BITS + TVN_BITS)) {
                index = (t.borrow().expires >> TVR_BITS) & (TVN_MASK as i64);
                vec = &mut self.tv2[index as usize];
            } else if idx < (1 << (TVR_BITS + 2 * TVN_BITS)) {
                index = (t.borrow().expires >> (TVR_BITS + 1 * TVN_BITS)) & (TVN_MASK as i64);
                vec = &mut self.tv3[index as usize];
            } else if idx < (1 << (TVR_BITS + 3 * TVN_BITS)) {
                index = (t.borrow().expires >> (TVR_BITS + 2 * TVN_BITS)) & (TVN_MASK as i64);
                vec = &mut self.tv4[index as usize];
            } else {
                index = (t.borrow().expires >> (TVR_BITS + 3 * TVN_BITS)) & (TVN_MASK as i64);
                vec = &mut self.tv5[index as usize];
            }
            crate::ListHead::add_tail(&vec,
                                      &t.borrow_mut().timer_entry);
        }
        return Rc::downgrade(&t.borrow().wp);
    }

    pub fn set_timeout(&mut self, time: i32, cb: TimerCb) -> TimerWp {
        return self.internal_add(TimerType::Timeout, time, cb, None);
    }

    pub fn set_timeout_ex
        (&mut self, time: i32, cb: TimerCb, attacher: TimerAttacherRp) -> TimerWp {
        return self.internal_add(TimerType::Timeout, time, cb, Some(attacher));
    }

    pub fn set_interval(&mut self, time: i32, cb: TimerCb) -> TimerWp {
        return self.internal_add(TimerType::Interval, time, cb, None);
    }

    pub fn set_interval_ex
        (&mut self, time: i32, cb: TimerCb, attacher: TimerAttacherRp) -> TimerWp {
        return self.internal_add(TimerType::Interval, time, cb, Some(attacher));
    }

    pub fn fire_event(&mut self,
                      _timer_wp: TimerWp,
                      _e: i32,
                      _args: Option<Vec<Rc::<dyn Any>>>) {
    }

    pub fn modify(&mut self, _timer_wp: TimerWp, _expire_time: i32) {
    }

    pub fn delete_current_timer(&mut self) {
    }

    pub fn is_running(&self) {
    }

    pub fn delete(&mut self, _timer_wp: TimerWp) {
    }

    pub fn reset(&mut self, _timer_wp: TimerWp) {
    }

    pub fn get_remain_time(&mut self, _timer_wp: TimerWp) -> i64 {
        return 0;
    }

    pub fn get_idle_time(&self) -> i64 {
        return 0;
    }

    fn gc_timer(&mut self, e: TimerEvent, _args: Option<Vec<Rc::<dyn Any>>>) {
        match e {
            TimerEvent::Exec => {

            }
            TimerEvent::Custom(_v) => {

            }
            _ => {

            }
        }
    }

}
